//Tento program po spusteni alokuje pamat pre GDT (32 poloziek), vyplni
//register GDTR, vytvori 5 deskriptorov pre CS,DS,ES,SS a video pamat
//a prepne procesor do chraneneho rezimu. Potom z neho vypise retazec,
//prepne procesor spat do realneho rezimu a program sa ukonci.
#include <dos.h>
#include <stdio.h>

//Zapneme instrukcie procesoru 386, zarovnanie na bajty
#pragma option -3 -a-

//Makro na vypocet linearnej adresy
#define LINEAR(seg,ofs) (16*(unsigned long)(seg)+(unsigned)(ofs))

//Struktura registru GDTR a IDTR
struct DTR {
  unsigned int limit;
  unsigned long base;
};

//Struktura deskriptoru
struct DESC {
  unsigned int limit;
  unsigned int base0;
  unsigned char base1;
  unsigned int flags;
  unsigned char base2;
};

//Typ deskriptoru - kodovy alebo datovy
enum TYPE { CODE, DATA };

//Hodnota segmentu pre GDT
unsigned GDT;

//Register GDTR
DTR GDTR;

//Hodnoty segmentovych registrov
unsigned seg_CS,seg_DS,seg_ES,seg_SS;


//Funkcia get_mode ma navratovu hodnotu:
//0 - realny rezim
//1 - virtualny rezim
int get_mode()
{
  asm smsw ax
  asm and ax,1
  return(_AX);
}


//Funkcia set_descriptor vyplni polozku deskriptoru v GDT na pozicii udanej
//selektorom podla zadanych parametrov
void set_descriptor(unsigned long base,unsigned long limit,TYPE type,int selector)
{
  DESC far *desc=(DESC far*)MK_FP(GDT,selector);
  limit=(limit-1)>>12;
  desc->limit=limit;
  desc->base0=base;
  desc->base1=base>>16;
  desc->base2=base>>24;
  switch (type) {
    case CODE: desc->flags=(limit & 0x0F0000)>>8 | 0x809A; break;
    case DATA: desc->flags=(limit & 0x0F0000)>>8 | 0x8092; }
}


//Funkcia show_message zobrazi v chranenom rezime spravu na obrazovku
void near show_message(char near *msg)
{ asm {
  push si di ds
  mov si,[msg]
  xor di,di
  mov ax,40
  mov es,ax                            //ES=selektor videopamati
  push ss
  pop ds }                             //DS=selektor zasobniku
  cyklus:
  asm {
    lodsb                              //AL=znak
    mov ah,0x1F                        //AH=farebny atribut
    or al,al                           //retazec konci nulou
    je koniec
    stosw
    jmp cyklus }
  koniec:
  asm pop ds di si
}


//Funkcia go_PM prepne procesor do chraneneho rezimu
void near go_PM()
{ asm {
  cli                                  //IF=0
  mov [seg_CS],cs
  mov [seg_DS],ds
  mov [seg_ES],es
  mov [seg_SS],ss                      //ulozenie registrov
  mov eax,cr0
  or ax,1                              //PE=1
  mov cr0,eax
  lgdt pword ptr [GDTR]
  mov ax,8
  lea dx,entryPM
  push ax                              //push CS selektor
  push dx                              //push IP
  retf
  entryPM:
  mov ax,16                            //DS selektor
  mov ds,ax
  mov ax,24                            //ES selektor
  mov es,ax
  mov ax,32                            //SS selektor
  mov ss,ax
  xor ax,ax                            //neplatny selektor
  mov fs,ax
  mov gs,ax
} }


//Funkcia go_RM prepne procesor do realneho rezimu
void near go_RM()
{ asm {
  cli                                  //IF=0
  mov eax,cr0
  and ax,0xFFFE                        //PE=0
  mov cr0,eax
  lea dx,entryRM
  push [seg_CS]                        //push CS segment
  push dx                              //push IP
  retf
  entryRM:
  xor ax,ax
  mov ds,[seg_DS]
  mov es,[seg_ES]
  mov ss,[seg_SS]                      //obnovenie registrov
  mov fs,ax
  mov gs,ax
  sti                                  //IF=1
} }


//Zaciatok programu
int main()
{
  //Zistime rezim procesoru
  if (!get_mode())
    printf("Procesor sa nachadza v realnom rezime.\n");
  else {
    printf("Procesor sa nachadza vo virtualnom rezime. Nemozno pokracovat.\n");
    return(0); }
  //Alokujeme 16 paragrafov (256 bajtov) pre GDT
  if (allocmem(16,&GDT)==-1)
    printf("Pamatovy blok pre GDT bol uspesne alokovany.\n");
  else {
    printf("Pamatovy blok pre GDT sa nepodarilo alokovat.\n");
    return(0); }
  //Vyplnime hodnotu registru GDTR
  GDTR.limit=0xFF;
  GDTR.base=LINEAR(GDT,0);
  printf("Register GDTR bol vyplneny.\n");
  //Vyplnime 5 deskriptorov podla hodnot segmentovych registrov
  //a deskriptor s bazovou adresou videopamati 0xB8000 (text)
  set_descriptor(LINEAR(_CS,0),0x10000,CODE,8);
  set_descriptor(LINEAR(_DS,0),0x10000,DATA,16);
  set_descriptor(LINEAR(_ES,0),0x10000,DATA,24);
  set_descriptor(LINEAR(_SS,0),0x10000,DATA,32);
  set_descriptor(0xB8000,0x10000,DATA,40);
  printf("Deskriptory boli uspesne vyplnene.\n");
  //Vstup do chraneneho rezimu
  go_PM();
  show_message("Chraneny rezim!");
  //Vstup do realneho rezimu
  go_RM();
  printf("Procesor je opat v realnom rezime.\n");
  return(0);
}